"""
RasgoQL Custom Errors
"""

class RasgoQLWarning(Warning):
    """
    Base warning attributable to RasgoQL code
    """
    pass

class RasgoQLException(Exception):
    """
    Base error attributable to RasgoQL code
    """
    pass

class DWCredentialsWarning(RasgoQLException):
    """
    Missing DataWarehouse credentials
    """
    pass

class DWConnectionError(RasgoQLException):
    """
    Error from DataWarehouse connection activity
    """
    pass

class DWQueryError(RasgoQLException):
    """
    Error from DataWarehouse query activity
    """
    pass

class ParameterException(RasgoQLException):
    """
    Problem populating a value-restricted parameter
    """
    pass

class PackageDependencyWarning(RasgoQLException):
    """
    Problem importing a required package dependency
    """
    pass

class SQLWarning(RasgoQLException):
    """
    Possible problem with SQL text
    """
    pass

class TableAccessError(RasgoQLException):
    """
    Error accessing Table
    """
    pass

class TableConflictException(RasgoQLException):
    """
    Possiblle conflict with Table names
    """
    pass

class TransformRenderingError(RasgoQLException):
    """
    Error from Transform rendering activity
    """
    pass



"""
Decorators
"""
import functools
import logging
from typing import Callable, Union

from rasgoql.primitives.enums import TableState

logging.basicConfig()
logger = logging.getLogger('rasgoQL')
logger.setLevel(logging.INFO)


def parametrized(decorator):
    """
    Meta-decorator to support passing parameters into future decorators
    """
    @functools.wraps(decorator)
    def decorator_maker(*args, **kwargs):
        def decorator_wrapper(func):
            return decorator(func, *args, **kwargs)
        return decorator_wrapper
    return decorator_maker

def beta(func: Callable) -> Callable:
    """
    Decorator to warn about beta features
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger.info(
            f'{func.__name__} is a beta feature. '
            'Its functionality and parameters may change in future versions and '
            'there may still be bugs to discover. If you experience any unexpected '
            'behaviors or have ideas to improve this function, please raise an issue '
            'in our github repo (https://github.com/rasgointelligence/RasgoQL/issues) '
            'or contact us directly on slack.'
        )
        return func(*args, **kwargs)
    return wrapper

def only_rasgo_tables(func: Callable) -> Callable:
    """
    Decorator to restrict Dataset methods
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self: 'Dataset' = args[0]
        if not self.is_rasgo:
            raise NotImplementedError(f'{func.__name__} method is only available for Rasgo-created tables and views')
        return func(*args, **kwargs)
    return wrapper

def require_dw(func: Callable) -> Callable:
    """
    Decorator to restrict Class methods
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self: Union['Dataset', 'Transform', 'SQLChain'] = args[0]
        if not self._dw:
            raise NotImplementedError(f'{func.__name__} method is only available for classes instantiated with a DW connection')
        return func(*args, **kwargs)
    return wrapper

def require_materialized(func: Callable) -> Callable:
    """
    Decorator to restrict Class methods
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self: 'Dataset' = args[0]
        if self.table_state != TableState.IN_DW.value:
            raise NotImplementedError(f'{func.__name__} method is only available for tables that exist in the DataWarehouse')
        return func(*args, **kwargs)
    return wrapper

def require_transforms(func: Callable) -> Callable:
    """
    Decorator to restrict Class methods
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self: 'SQLChain' = args[0]
        if not self.transforms:
            raise NotImplementedError(f'{func.__name__} method is only available for SQLChains with at least one Transform')
        return func(*args, **kwargs)
    return wrapper

@parametrized
def restrict_table_type(func: Callable, table_type: str) -> Callable:
    """
    Decorator to restrict Dataset methods
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self: 'Dataset' = args[0]
        if self.table_type != table_type.upper():
            raise NotImplementedError(f'{func.__name__} method is only available for {table_type.upper()}s')
        return func(*args, **kwargs)
    return wrapper



"""
Verbose messaging functions
"""
import logging

RQL_RUN_VERBOSE: bool = False

def verbose_message(
        message: str,
        logger: logging.Logger
    ) -> None:
    """
    Logs messages only if verbose flag is set to true
    """
    logger.setLevel(logging.INFO)
    if RQL_RUN_VERBOSE:
        logger.info(message)

def set_verbose(
        value: bool
    ) -> None:
    """
    Set global var
    """
    global RQL_RUN_VERBOSE
    RQL_RUN_VERBOSE = value



"""
BigQuery Data Warehouse classes
"""

import logging
import os
from typing import List, Tuple, Union

import json
import pandas as pd

from rasgoql.data.base import DataWarehouse, DWCredentials
from rasgoql.errors import (
    DWCredentialsWarning, DWConnectionError, DWQueryError,
    ParameterException, PackageDependencyWarning,
    SQLWarning, TableAccessError, TableConflictException
)
from rasgoql.imports import bq, gcp_exc, gcp_flow, gcp_svc
from rasgoql.primitives.enums import (
    check_response_type,
    check_write_method, check_write_table_type
)
from rasgoql.utils.creds import load_env, save_env
from rasgoql.utils.df import cleanse_sql_dataframe
from rasgoql.utils.messaging import verbose_message
from rasgoql.utils.sql import (
    is_scary_sql, magic_fqtn_handler,
    parse_fqtn, parse_namespace,
    validate_namespace
)

logging.basicConfig()
logger = logging.getLogger('BigQuery DataWarehouse')
logger.setLevel(logging.INFO)


class BigQueryCredentials(DWCredentials):
    """
    BigQuery Credentials
    """
    dw_type = 'bigquery'

    def __init__(
            self,
            json_filepath: str,
            project: str = None,
            dataset: str = None
        ):
        self.json_filepath = json_filepath
        self.project = project
        self.dataset = dataset

    def __repr__(self) -> str:
        return json.dumps(
            {
                "json_filepath": self.json_filepath,
                "project": self.project,
                "dataset": self.dataset
            }
        )

    @classmethod
    def from_env(
            cls,
            filepath: str = None
        ) -> 'BigQueryCredentials':
        """
        Creates an instance of this Class from a .env file on your machine
        """
        load_env(filepath)
        json_filepath = os.getenv('BIGQUERY_JSON_FILEPATH', os.getenv('GOOGLE_APPLICATION_CREDENTIALS'))
        project = os.getenv('BIGQUERY_PROJECT')
        dataset = os.getenv('BIGQUERY_DATASET')
        if not all([json_filepath, project, dataset]):
            raise DWCredentialsWarning(
                'Your env file is missing expected credentials. Consider running '
                'BigQueryCredentials(*args).to_env() to repair this.'
            )
        return cls(
            json_filepath,
            project,
            dataset
        )

    def to_dict(self) -> dict:
        """
        Returns a dict of the credentials
        """
        return {
            "json_filepath": self.json_filepath,
            "project": self.project,
            "dataset": self.dataset
        }

    def to_env(
            self,
            filepath: str = None,
            overwrite: bool = False
        ):
        """
        Saves credentials to a .env file on your machine
        """
        creds = {
            "BIGQUERY_JSON_FILEPATH": self.json_filepath,
            "BIGQUERY_PROJECT": self.project,
            "BIGQUERY_DATASET": self.dataset,
        }
        return save_env(creds, filepath, overwrite)


class BigQueryDataWarehouse(DataWarehouse):
    """
    Google BigQuery DataWarehouse
    """
    dw_type = 'bigquery'
    credentials_class = BigQueryCredentials

    def __init__(self):
        if bq is None:
            raise PackageDependencyWarning(
                'Missing a required python package to run BigQuery. '
                'Please download the BigQuery package by running: '
                'pip install rasgoql[bigquery]')

        super().__init__()
        self.credentials: dict = None
        self.connection: bq.Client = None
        self.default_project = None
        self.default_dataset = None

    # ---------------------------
    # Core Data Warehouse methods
    # ---------------------------
    def change_namespace(
            self,
            namespace: str
        ) -> None:
        """
        Changes the default namespace of your connection

        Params:
        `namespace`: str:
            namespace (project.dataset)
        """
        namespace = self._validate_namespace(namespace)
        project, dataset = parse_namespace(namespace)
        try:
            self.execute_query(f'USE PROJECT {project}')
            self.execute_query(f'USE DATASET {dataset}')
            self.default_namespace = namespace
            self.default_project = project
            self.default_dataset = dataset
            verbose_message(
                f"Namespace reset to {self.default_namespace}",
                logger
            )
        except Exception as e:
            self._error_handler(e)

    def connect(
            self,
            credentials: Union[dict, BigQueryCredentials]
        ):
        """
        Connect to BigQuery

        Params:
        `credentials`: dict:
            dict (or DWCredentials class) holding the connection credentials
        """
        if isinstance(credentials, BigQueryCredentials):
            credentials = credentials.to_dict()

        try:
            self.default_project = credentials.get('project')
            self.default_dataset = credentials.get('dataset')
            if not self.credentials:
                self.credentials = self._get_credentials(
                    credentials.get('json_filepath')
                )
            self.connection = bq.Client(
                credentials=self.credentials,
                project=self.default_project
            )
            verbose_message(
                "Connected to BigQuery",
                logger
            )
        except Exception as e:
            self._error_handler(e)

    def close_connection(self):
        """
        Close connection to BigQuery
        """
        try:
            if self.connection:
                self.connection.close()
            self.connection = None
            verbose_message(
                "Connection to BigQuery closed",
                logger
            )
        except Exception as e:
            self._error_handler(e)

    def create(
            self,
            sql: str,
            fqtn: str,
            table_type: str = 'VIEW',
            overwrite: bool = False
        ):
        """
        Create a view or table from given SQL

        Params:
        `sql`: str:
            query that returns data (i.e. can be wrapped in a CREATE TABLE statement)
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
            Name for the new table
        `table_type`: str:
            One of values: [view, table]
        `overwrite`: bool
            pass True when this table name already exists in your DataWarehouse
            and you know you want to overwrite it
            WARNING: This will completely overwrite data in the existing table
        """
        table_type = check_write_table_type(table_type)
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        if self._table_exists(fqtn) and not overwrite:
            msg = f'A table or view named {fqtn} already exists. ' \
                   'If you are sure you want to overwrite it, ' \
                   'pass in overwrite=True and run this function again'
            raise TableConflictException(msg)
        query = f'CREATE OR REPLACE {table_type} {fqtn} AS {sql}'
        self.execute_query(query, acknowledge_risk=True, response='None')
        return fqtn

    @property
    def default_namespace(self) -> str:
        """
        Returns the default project.dataset of this connection
        """
        return f'{self.default_project}.{self.default_dataset}'

    def execute_query(
            self,
            sql: str,
            response: str = 'tuple',
            acknowledge_risk: bool = False
        ):
        """
        Run a query against BigQuery and return all results

        `sql`: str:
            query text to execute
        `response`: str:
            Possible values: [dict, df, None]
        `acknowledge_risk`: bool:
            pass True when you know your SQL statement contains
            a potentially dangerous or data-altering operation
            and still want to run it against your DataWarehouse
        """
        response = check_response_type(response)
        if is_scary_sql(sql) and not acknowledge_risk:
            msg = 'It looks like your SQL statement contains a ' \
                  'potentially dangerous or data-altering operation.' \
                  'If you are positive you want to run this, ' \
                  'pass in acknowledge_risk=True and run this function again.'
            raise SQLWarning(msg)
        verbose_message(
            f"Executing query: {sql}",
            logger
        )
        if response == 'DICT':
            return self._query_into_dict(sql)
        if response == 'DF':
            return self._query_into_pandas(sql)
        return self._execute_string(sql, ignore_results=(response == 'NONE'))

    def get_ddl(
            self,
            fqtn: str
        ) -> str:
        """
        Returns the create statement for a table or view

        `fqtn`: str:
            Fully-qualified Table Name (database.schema.table)
        """
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        proj, ds, tbl = parse_fqtn(fqtn)
        sql = f"SELECT DDL FROM {proj}.{ds}.INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME='{tbl}'"
        query_response = self.execute_query(sql)
        return query_response[0]

    def get_object_details(
            self,
            fqtn: str
        ) -> tuple:
        """
        Return details of a table or view in BigQuery

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)

        Response:
            object exists: bool
            is rasgo object: bool
            object type: [table|view|unknown]
        """
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        obj_exists = False
        is_rasgo_obj = False
        obj_type = 'unknown'
        try:
            table = self.connection.get_table(fqtn)
            obj_exists = True
            obj_type = table.table_type
            if table.labels:
                for _label, value in table.labels.items():
                    if value == 'rasgoql':
                        is_rasgo_obj = True
            return obj_exists, is_rasgo_obj, obj_type
        except gcp_exc.NotFound:
            return obj_exists, is_rasgo_obj, obj_type

    def get_schema(
            self,
            fqtn: str,
            create_sql: str = None
        ) -> Tuple[str, str]:
        """
        Return the schema of a table or view

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        `create_sql`: str:
            A SQL select statement that will create the view. If this param is passed
            and the fqtn does not already exist, it will be created and profiled based
            on this statement. The view will be dropped after profiling
        """
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        response = []
        try:
            if self._table_exists(fqtn):
                table = self.connection.get_table(fqtn)
            elif create_sql:
                self.create(create_sql, fqtn, table_type='view')
                table = self.connection.get_table(fqtn)
                self.execute_query(f'DROP VIEW {fqtn}', response='none', acknowledge_risk=True)
            else:
                raise TableAccessError(f'Table {fqtn} does not exist or cannot be accessed.')
            for schema_field in table.schema:
                response.append((schema_field.name, schema_field.field_type))
            return response
        except Exception as e:
            self._error_handler(e)

    def list_tables(
            self,
            database: str = None,
            schema: str = None
        ) -> pd.DataFrame:
        """
        List all tables and views available in default namespace

        Params:
        `database`: str:
            override database
        `schema`: str:
            override schema
        """
        project = database or self.default_project
        dataset = schema or self.default_dataset
        namespace = f'{project}.{dataset}'
        try:
            tables = self.connection.list_tables(namespace)
            records = []
            columns = [
                'TABLE_NAME',
                'FQTN',
                'TABLE_TYPE',
                'ROW_COUNT',
                'CREATED',
                'LAST_ALTERED'
            ]
            for tbl in tables:
                table = self.connection.get_table(tbl)
                records.append(
                    (
                        table.table_id,
                        f'{table.project}.{table.dataset_id}.{table.table_id}',
                        table.table_type,
                        table.num_rows,
                        table.created,
                        table.modified
                    )
                )
            return pd.DataFrame(
                records,
                columns=columns
            )
        except Exception as e:
            self._error_handler(e)

    def preview(
            self,
            sql: str,
            limit: int = 10
        ) -> pd.DataFrame:
        """
        Returns 10 records into a pandas DataFrame

        Params:
        `sql`: str:
            SQL statment to run
        `limit`: int:
            Records to return
        """
        return self.execute_query(
            f'{sql} LIMIT {limit}',
            response='df',
            acknowledge_risk=True
        )

    def save_df(
            self,
            df: pd.DataFrame,
            fqtn: str,
            method: str = None
        ) -> str:
        """
        Creates a table in BigQuery from a pandas Dataframe

        Params:
        `df`: pandas DataFrame:
            DataFrame to upload
        `fqtn`: str:
            Fully-qualied table name (database.schema.table)
            Name for the new table
        `overwrite`: bool
            pass True when this table name already exists in your DataWarehouse
            and you know you want to overwrite it
            WARNING: This will completely overwrite data in the existing table
        """
        if method:
            method = check_write_method(method)
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        table_exists = self._table_exists(fqtn)
        if table_exists and not method:
            msg = f"A table named {fqtn} already exists. " \
                   "If you are sure you want to write over it, pass in " \
                   "method='append' or method='replace' and run this function again"
            raise TableConflictException(msg)
        try:
            cleanse_sql_dataframe(df)
            job_config = bq.LoadJobConfig(
                write_disposition='WRITE_TRUNCATE' if method == 'REPLACE' else None
            )
            _job = self.connection.load_table_from_dataframe(
                df,
                fqtn,
                job_config=job_config
            )
            # Wait for the job to complete
            _job.result()
            return fqtn
        except Exception as e:
            self._error_handler(e)


    # ---------------------------
    # Core Data Warehouse helpers
    # ---------------------------
    def _table_exists(
            self,
            fqtn: str
        ) -> bool:
        """
        Check for existence of fqtn in the Data Warehouse and return a boolean

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        """
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        try:
            self.connection.get_table(fqtn)
            return True
        except gcp_exc.NotFound:
            return False

    def _validate_namespace(
            self,
            namespace: str
        ) -> str:
        """
        Checks a namespace string for compliance with BigQuery format

        Params:
        `namespace`: str:
            namespace (project.dataset)
        """
        # Does this match a 'string.string' pattern?
        try:
            validate_namespace(namespace)
            return namespace
        except ValueError:
            raise ParameterException("Bigquery namespaces should be format: PROJECT.DATASET")

    # --------------------------
    # BigQuery specific helpers
    # --------------------------
    @property
    def _default_job_config(self) -> 'bq.QueryJobConfig':
        return bq.QueryJobConfig(
            default_dataset=self.default_namespace
            )

    def _error_handler(
            self,
            exception: Exception,
            query: str = None
        ) -> None:
        """
        Handle Snowflake exceptions that need additional info
        """
        verbose_message(
            f'Exception occurred while running query: {query}',
            logger
        )
        if exception is None:
            return
        if isinstance(exception, gcp_exc.NotFound):
            raise TableAccessError('Table does not exist') from exception
        if isinstance(exception, gcp_exc.BadRequest):
            if {"reason": "invalidQuery"} in exception.errors:
                raise DWQueryError(exception.errors) from exception
        if isinstance(exception, gcp_exc.ServiceUnavailable):
            raise DWConnectionError(
                'BigQuery is unavailable. Please check that your are using '
                'valid credentials, that you have internet access, and '
                'that http connections to GCP are whitelisted in your env. '
                'Finally check https://status.cloud.google.com/incident/bigquery '
                'for outage status.'
            ) from exception
        raise exception

    def _execute_string(
            self,
            query: str,
            ignore_results: bool = False
        ) -> List[tuple]:
        """
        Execute a query string against the DataWarehouse connection and fetch all results
        """
        try:
            query_job = self.connection.query(
                query,
                job_config=self._default_job_config
            )
            if ignore_results:
                return
            return list(query_job.result())
        except Exception as e:
            self._error_handler(e)

    def _get_appflow_credentials(
            self,
            filepath: str
        ) -> 'bq.Credentials':
        try:
            appflow = gcp_flow.InstalledAppFlow.from_client_secrets_file(
                filepath,
                scopes=["https://www.googleapis.com/auth/bigquery"]
            )
            appflow.run_local_server()
            #appflow.run_console()
            return appflow.credentials
        except Exception as e:
            self._error_handler(e)

    def _get_credentials(
            self,
            filepath: str
        ) -> 'bq.Credentials':
        try:
            return self._get_appflow_credentials(filepath)
        except ValueError:
            return self._get_service_credentials(filepath)

    def _get_service_credentials(
            self,
            filepath: str
        ) -> 'bq.Credentials':
        try:
            credentials = gcp_svc.Credentials.from_service_account_file(
                filepath,
                scopes=["https://www.googleapis.com/auth/cloud-platform"]
            )
            return credentials
        except Exception as e:
            self._error_handler(e)

    def _query_into_dict(
            self,
            query: str
    ) -> dict:
        """
        Return results of query in a pandas DataFrame
        """
        try:
            return self.connection.query(
                query,
                job_config=self._default_job_config
                ) \
                .result() \
                .to_dataframe() \
                .to_dict()
        except Exception as e:
            self._error_handler(e)

    def _query_into_pandas(
            self,
            query: str
    ) -> pd.DataFrame:
        """
        Return results of query in a pandas DataFrame
        """
        try:
            return self.connection.query(
                query,
                job_config=self._default_job_config
                ) \
                .result() \
                .to_dataframe()
        except Exception as e:
            self._error_handler(e)



"""
Holds dbt helpers
"""
import logging
import os
import string
from typing import List, Tuple

from pathlib import Path

import yaml

from rasgoql.utils.sql import parse_namespace

logging.basicConfig()
logger = logging.getLogger('dbt')
logger.setLevel(logging.INFO)

DBT_MODEL_CONFIG_TEMPLATE = '''
  config(
    {config_args}
  )
'''

DBT_PROJECT_TEMPLATE = {
    "name": "",
    "version": "1.0.0",
    "config-version": 2,
    "profile": "default",
    "model-paths": ["models"],
    "analysis-paths": ["analyses"],
    "test-paths": ["tests"],
    "seed-paths": ["seeds"],
    "macro-paths": ["macros"],
    "snapshot-paths": ["snapshots"],
    "target-path": "target",
    "log-path": "logs",
    "packages-install-path": "dbt_packages",
    "clean-targets": ["target", "dbt_packages"],
    "models": None
}


def check_project_name(
        project_name: str
    ):
    """
    Checks a project name for dbt compliance
    """
    allowed = set(string.ascii_lowercase + '_')
    if any(char for char in project_name.lower() if char not in allowed):
        logger.warning(
            "per dbt: Project names should contain only lowercase characters "
            "and underscores. A good package name should reflect your organization's "
            "name or the intended use of these models"
        )
    return project_name.lower()

def prepare_dbt_path(
        project_name: str,
        project_directory: Path
    ) -> None:
    """
    Checks for a specified filepath and creates one if it doesn't exist
    """
    dbt_dirs = ['analyses', 'dbt_packages', 'logs', 'macros',
                'models', 'seeds', 'target', 'tests']
    if project_name not in project_directory:
        project_directory = os.path.join(project_directory, project_name)
    if not os.path.exists(project_directory):
        os.makedirs(project_directory)
    for dir_name in dbt_dirs:
        this_dir = os.path.join(project_directory, dir_name)
        if not os.path.exists(this_dir):
            os.makedirs(this_dir)
    return project_directory

def save_project_file(
        project_name: str,
        filepath: Path,
        namespace: str,
        materialize: str
    ) -> bool:
    """
    Writes a yaml definition to a dbt project file
    """
    if not os.path.exists(filepath):
        yml_definition = DBT_PROJECT_TEMPLATE.copy()
        yml_definition["name"] = project_name
        model_config = {project_name: {"+materialize": materialize}}
        if namespace:
            db, schema = parse_namespace(namespace)
            model_config[project_name].update({"database": db})
            model_config[project_name].update({"schema": schema})
        yml_definition["models"] = model_config
        with open(filepath, "w") as _yaml:
            yaml.dump(data=yml_definition, Dumper=yaml.SafeDumper, stream=_yaml, sort_keys=False)
    return filepath

def save_model_file(
        sql_definition: str,
        output_directory: Path,
        file_name: str,
        schema: List[Tuple[str, str]],
        config_args: dict = None,
        include_schema: bool = False
) -> bool:
    """
    Writes a sql script to a dbt model file
    """
    filepath = os.path.join(output_directory, file_name)
    if config_args:
        # TODO: Should we validate whether args are valid dbt keywords?
        model_config = DBT_MODEL_CONFIG_TEMPLATE.format(
            config_args=config_args
        )
        model_config = '{{' + model_config + '}}'
        sql_definition = f'{model_config}\n\n{sql_definition}'
    with open(filepath, "w") as _file:
        _file.write(sql_definition)
    if include_schema:
        model_name = file_name.replace('.sql', '')
        save_schema_file(output_directory, model_name, schema, config_args)
    return filepath

def save_schema_file(
        output_directory: Path,
        model_name: str,
        schema: List[Tuple[str, str]],
        config_args: dict = None,
    ):
    """
    Writes a table def to a dbt schema file
    """
    filepath = os.path.join(output_directory, 'schema.yml')
    schema_definition = None
    columns_list = []
    for row in schema:
        columns_list.append({"name:": row[0]})
    model_dict = {"name": model_name, "columns": columns_list}
    if config_args:
        model_dict.update({"config": config_args})
    if not os.path.exists(filepath):
        schema_definition = {"version": 2, "models": [model_dict]}
    else:
        schema_definition = [model_dict]
    with open(filepath, "a") as _file:
        yaml.dump(data=schema_definition, Dumper=yaml.SafeDumper, stream=_file, sort_keys=False)
    return filepath



from rasgoql.data.base import DataWarehouse, DWCredentials
from rasgoql.data.bigquery import BigQueryDataWarehouse, BigQueryCredentials
from rasgoql.data.snowflake import SnowflakeDataWarehouse, SnowflakeCredentials
from rasgoql.data.postgres import PostgresCredentials, PostgresDataWarehouse



"""
Snowflake DataWarehouse classes
"""
import logging
import os
from typing import List, Optional, Union

import json
import pandas as pd

from rasgoql.data.base import DataWarehouse, DWCredentials
from rasgoql.errors import (
    DWCredentialsWarning, DWConnectionError, DWQueryError,
    PackageDependencyWarning, ParameterException,
    SQLWarning, TableAccessError, TableConflictException
)
from rasgoql.imports import sf_connector, write_pandas
from rasgoql.primitives.enums import (
    check_response_type, check_write_method, check_write_table_type
)
from rasgoql.utils.creds import load_env, save_env
from rasgoql.utils.df import cleanse_sql_dataframe, generate_dataframe_ddl
from rasgoql.utils.messaging import verbose_message
from rasgoql.utils.sql import (
    is_scary_sql, magic_fqtn_handler,
    parse_fqtn, parse_namespace,
    validate_namespace
)

logging.basicConfig()
logger = logging.getLogger('Snowflake DataWarehouse')
logger.setLevel(logging.INFO)


class SnowflakeCredentials(DWCredentials):
    """
    Snowflake Credentials
    """
    dw_type = 'snowflake'

    def __init__(
            self,
            account: str,
            user: str,
            password: str,
            role: str,
            warehouse: str,
            database: str,
            schema: str
        ):
        if sf_connector is None:
            raise PackageDependencyWarning(
                'Missing a required python package to run Snowflake. '
                'Please download the Snowflake package by running: '
                'pip install rasgoql[snowflake]')
        self.account = account
        self.user = user
        self.password = password
        self.role = role
        self.warehouse = warehouse
        self.database = database
        self.schema = schema

    def __repr__(self) -> str:
        return json.dumps(
            {
                "account": self.account,
                "user": self.user,
                "role": self.role,
                "database": self.database,
                "schema": self.schema,
                "warehouse": self.warehouse,
            }
        )

    @classmethod
    def from_env(
            cls,
            filepath: str = None
        ) -> 'SnowflakeCredentials':
        """
        Creates an instance of this Class from a .env file on your machine
        """
        load_env(filepath)
        account = os.getenv('SNOWFLAKE_ACCOUNT')
        user = os.getenv('SNOWFLAKE_USER')
        password = os.getenv('SNOWFLAKE_PASSWORD')
        role = os.getenv('SNOWFLAKE_ROLE')
        warehouse = os.getenv('SNOWFLAKE_WAREHOUSE')
        database = os.getenv('SNOWFLAKE_DATABASE')
        schema = os.getenv('SNOWFLAKE_SCHEMA')
        if not all([account, user, password, role, warehouse, database, schema]):
            raise DWCredentialsWarning(
                'Your env file is missing expected credentials. Consider running '
                'SnowflakeCredentials(*args).to_env() to repair this.'
            )
        return cls(
            account,
            user,
            password,
            role,
            warehouse,
            database,
            schema
        )

    def to_dict(self) -> dict:
        """
        Returns a dict of the credentials
        """
        return {
            "account": self.account,
            "user": self.user,
            "password": self.password,
            "role": self.role,
            "warehouse": self.warehouse,
            "database": self.database,
            "schema": self.schema
        }

    def to_env(
            self,
            filepath: str = None,
            overwrite: bool = False
        ):
        """
        Saves credentials to a .env file on your machine
        """
        creds = {
            "SNOWFLAKE_ACCOUNT": self.account,
            "SNOWFLAKE_USER": self.user,
            "SNOWFLAKE_PASSWORD": self.password,
            "SNOWFLAKE_ROLE": self.role,
            "SNOWFLAKE_WAREHOUSE": self.warehouse,
            "SNOWFLAKE_DATABASE": self.database,
            "SNOWFLAKE_SCHEMA": self.schema,
        }
        return save_env(creds, filepath, overwrite)


class SnowflakeDataWarehouse(DataWarehouse):
    """
    Snowflake DataWarehouse
    """
    dw_type = 'snowflake'
    credentials_class = SnowflakeCredentials

    def __init__(self):
        super().__init__()
        self.credentials: dict = None
        self.connection: sf_connector.SnowflakeConnection = None
        self.default_database = None
        self.default_schema = None

    # ---------------------------
    # Core Data Warehouse methods
    # ---------------------------
    def change_namespace(
            self,
            namespace: str
        ) -> None:
        """
        Changes the default namespace of your connection

        Params:
        `namespace`: str:
            namespace (database.schema)
        """
        namespace = self._validate_namespace(namespace)
        database, schema = parse_namespace(namespace)
        try:
            self.execute_query(f'USE DATABASE {database}')
            self.execute_query(f'USE SCHEMA {schema}')
            self.default_namespace = namespace
            self.default_database = database
            self.default_schema = schema
            verbose_message(
                f"Namespace reset to {self.default_namespace}",
                logger
            )
        except Exception as e:
            self._error_handler(e)

    def connect(
            self,
            credentials: Union[dict, SnowflakeCredentials]
        ):
        """
        Connect to Snowflake

        Params:
        `credentials`: dict:
            dict (or DWCredentials class) holding the connection credentials
        """
        if isinstance(credentials, SnowflakeCredentials):
            credentials = credentials.to_dict()

        # This allows you to track what queries were run by RasgoQL in your history tab
        credentials.update({
            "application": "rasgoql",
            "session_parameters": {
                "QUERY_TAG": "rasgoql"
            }
        })
        try:
            self.credentials = credentials
            self.default_database = credentials.get('database')
            self.default_schema = credentials.get('schema')
            self.connection = sf_connector.connect(**credentials)
            verbose_message(
                "Connected to Snowflake",
                logger
            )
        except Exception as e:
            self._error_handler(e)

    def close_connection(self):
        """
        Close connection to Snowflake
        """
        try:
            if self.connection:
                self.connection.close()
            self.connection = None
            verbose_message(
                "Connection to Snowflake closed",
                logger
            )
        except Exception as e:
            self._error_handler(e)

    def create(
            self,
            sql: str,
            fqtn: str,
            table_type: str = 'VIEW',
            overwrite: bool = False
        ):
        """
        Create a view or table from given SQL

        Params:
        `sql`: str:
            query that returns data (i.e. can be wrapped in a CREATE TABLE statement)
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
            Name for the new table
        `table_type`: str:
            One of values: [view, table]
        `overwrite`: bool
            pass True when this table name already exists in your DataWarehouse
            and you know you want to overwrite it
            WARNING: This will completely overwrite data in the existing table
        """
        table_type = check_write_table_type(table_type)
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        if self._table_exists(fqtn) and not overwrite:
            msg = f'A table or view named {fqtn} already exists. ' \
                   'If you are sure you want to overwrite it, ' \
                   'pass in overwrite=True and run this function again'
            raise TableConflictException(msg)
        query = f"CREATE OR REPLACE {table_type} {fqtn} COMMENT='rasgoql' AS {sql}"
        self.execute_query(query, acknowledge_risk=True, response='None')
        return fqtn

    @property
    def default_namespace(self) -> str:
        """
        Returns the default database.schema of this connection
        """
        return f'{self.default_database}.{self.default_schema}'

    def execute_query(
            self,
            sql: str,
            response: str = 'tuple',
            acknowledge_risk: bool = False
        ):
        """
        Run a query against Snowflake and return all results

        `sql`: str:
            query text to execute
        `response`: str:
            Possible values: [dict, df, None]
        `acknowledge_risk`: bool:
            pass True when you know your SQL statement contains
            a potentially dangerous or data-altering operation
            and still want to run it against your DataWarehouse
        """
        response = check_response_type(response)
        if is_scary_sql(sql) and not acknowledge_risk:
            msg = 'It looks like your SQL statement contains a ' \
                  'potentially dangerous or data-altering operation.' \
                  'If you are positive you want to run this, ' \
                  'pass in acknowledge_risk=True and run this function again.'
            raise SQLWarning(msg)
        verbose_message(
            f"Executing query: {sql}",
            logger
        )
        if response == 'DICT':
            return self._execute_dict_cursor(sql)
        if response == 'DF':
            return self._execute_df_cursor(sql)
        return self._execute_string(sql, ignore_results=(response == 'NONE'))

    def get_ddl(
            self,
            fqtn: str
        ) -> str:
        """
        Returns the create statement for a table or view

        `fqtn`: str:
            Fully-qualified Table Name (database.schema.table)
        """
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        sql = f"SELECT GET_DDL('TABLE', '{fqtn}') AS DDL"
        query_response = self.execute_query(sql, response='dict')
        return query_response[0]['DDL']

    def get_object_details(
            self,
            fqtn: str
        ) -> tuple:
        """
        Return details of a table or view in Snowflake

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)

        Response:
            object exists: bool
            is rasgo object: bool
            object type: [table|view|unknown]
        """
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        database, schema, table = parse_fqtn(fqtn)
        sql = f"SHOW OBJECTS LIKE '{table}' IN {database}.{schema}"
        result = self.execute_query(sql, response='dict')
        obj_exists = len(result) > 0
        is_rasgo_obj = False
        obj_type = 'unknown'
        if obj_exists:
            is_rasgo_obj = (result[0].get('comment') == 'rasgoql')
            obj_type = result[0].get('kind')
        return obj_exists, is_rasgo_obj, obj_type

    def get_schema(
            self,
            fqtn: str,
            create_sql: str = None
        ) -> dict:
        """
        Return the schema of a table or view

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        `create_sql`: str:
            A SQL select statement that will create the view. If this param is passed
            and the fqtn does not already exist, it will be created and profiled based
            on this statement. The view will be dropped after profiling
        """
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        desc_sql = f"DESC TABLE {fqtn}"
        response = []
        try:
            if self._table_exists(fqtn):
                query_response = self.execute_query(desc_sql, response='dict')
            elif create_sql:
                self.create(create_sql, fqtn, table_type='view')
                query_response = self.execute_query(desc_sql, response='dict')
                self.execute_query(f'DROP VIEW {fqtn}', response='none', acknowledge_risk=True)
            else:
                raise TableAccessError(f'Table {fqtn} does not exist or cannot be accessed.')
            for row in query_response:
                response.append((row['name'], row['type']))
            return response
        except Exception as e:
            self._error_handler(e)

    def list_tables(
            self,
            database: str = None,
            schema: str = None
        ) -> pd.DataFrame:
        """
        List all tables and views available in default namespace

        Params:
        `database`: str:
            override database
        `schema`: str:
            override schema
        """
        select_clause = "SELECT TABLE_NAME, " \
                        "TABLE_CATALOG||'.'||TABLE_SCHEMA||'.'||TABLE_NAME AS FQTN, " \
                        "CASE TABLE_TYPE WHEN 'BASE TABLE' THEN 'TABLE' ELSE TABLE_TYPE END AS TABLE_TYPE, " \
                        "ROW_COUNT, CREATED, LAST_ALTERED "
        from_clause = " FROM INFORMATION_SCHEMA.TABLES "
        if database:
            from_clause = f" FROM {database.upper()}.INFORMATION_SCHEMA.TABLES "
        where_clause = f" WHERE TABLE_SCHEMA = '{schema.upper()}'" if schema else ""
        sql = select_clause + from_clause + where_clause
        return self.execute_query(sql, response='df', acknowledge_risk=True)

    def preview(
            self,
            sql: str,
            limit: int = 10
        ) -> pd.DataFrame:
        """
        Returns 10 records into a pandas DataFrame

        Params:
        `sql`: str:
            SQL statment to run
        `limit`: int:
            Records to return
        """
        return self.execute_query(
            f'{sql} LIMIT {limit}',
            response='df',
            acknowledge_risk=True
        )

    def save_df(
            self,
            df: pd.DataFrame,
            fqtn: str,
            method: str = None
        ) -> str:
        """
        Creates a table in Snowflake from a pandas Dataframe

        Params:
        `df`: pandas DataFrame:
            DataFrame to upload
        `fqtn`: str:
            Fully-qualied table name (database.schema.table)
            Name for the new table
        `method`: str
            Values: [append, replace]
            when this table already exists in your DataWarehouse,
            pass append: to add dataframe rows to it
            pass replace: to overwrite it with dataframe rows
                WARNING: This will completely overwrite data in the existing table
        """
        if method:
            method = check_write_method(method)
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        table_exists = self._table_exists(fqtn)
        if table_exists and not method:
            msg = f"A table named {fqtn} already exists. " \
                   "If you are sure you want to write over it, pass in " \
                   "method='append' or method='replace' and run this function again"
            raise TableConflictException(msg)
        try:
            cleanse_sql_dataframe(df)
            # If the table does not exist or we've received instruction to replace
            # Issue a create or replace statement before we insert data
            if not table_exists or method == 'REPLACE':
                create_stmt = generate_dataframe_ddl(df, fqtn)
                create_stmt += " COMMENT='rasgoql' "
                self.execute_query(create_stmt, response='None', acknowledge_risk=True)
            _success, _chunks, _rows, _output = write_pandas(
                conn=self.connection,
                df=df,
                table_name=fqtn,
                quote_identifiers=False
            )
            return fqtn
        except Exception as e:
            self._error_handler(e)

    # ---------------------------
    # Core Data Warehouse helpers
    # ---------------------------
    def _table_exists(
            self,
            fqtn: str
        ) -> bool:
        """
        Check for existence of fqtn in the Data Warehouse and return a boolean

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        """
        fqtn = magic_fqtn_handler(fqtn, self.default_namespace)
        do_i_exist, _, _ = self.get_object_details(fqtn)
        return do_i_exist

    def _validate_namespace(
            self,
            namespace: str
        ) -> str:
        """
        Checks a namespace string for compliance with Snowflake format

        Params:
        `namespace`: str:
            namespace (database.schema)
        """
        try:
            validate_namespace(namespace)
            return namespace.upper()
        except ValueError:
            raise ParameterException("Snowflake namespaces should be format: DATABASE.SCHEMA")

    # --------------------------
    # Snowflake specific helpers
    # --------------------------
    def _error_handler(
            self,
            exception: Exception,
            query: str = None
        ) -> None:
        """
        Handle Snowflake exceptions that need additional info
        """
        verbose_message(
            f"Exception occurred while running query: {query}",
            logger
        )
        if exception is None:
            return
        if isinstance(exception, sf_connector.errors.ProgrammingError):
            if exception.errno == 3001:
                raise TableAccessError(
                    'You do not have access to operate on this object. '
                    'Two possible ways to resolve: '
                    'Connect with different credentials that have the proper access. '
                    'Or run `.change_namespace` on your SQLChain to write it to a '
                    'namespace your credentials can access'
                ) from exception
        if isinstance(exception, sf_connector.errors.DatabaseError):
            if exception.errno == 250001:
                raise DWConnectionError(
                    'Invalid username / password, please check that your '
                    'credentials are correct and try to reconnect.'
                ) from exception
        if isinstance(exception, sf_connector.errors.ServiceUnavailableError):
            raise DWConnectionError(
                'Snowflake is unavailable. Please check that your are using '
                'a valid account identifier, that you have internet access, and '
                'that http connections to Snowflake are whitelisted in your env. '
                'Finally check https://status.snowflake.com/ for outage status.'
            ) from exception
        raise exception

    def _execute_dict_cursor(
            self,
            query: str
        ) -> List[dict]:
        """
        Run a query string and return results in a Snowflake DictCursor

        PRO:
        Results are callable by column name
        for row in data:
            row['COL_NAME']

        CON:
        Query string must be a single statement (only one ;) or Snowflake returns an error
        """
        cursor = None
        try:
            cursor = self.connection.cursor(sf_connector.DictCursor)
            query_return = cursor.execute(query).fetchall()
            return query_return
        except Exception as e:
            self._error_handler(e)
        finally:
            if cursor:
                cursor.close()

    def _execute_df_cursor(
            self,
            query: str,
            params: Optional[dict] = None
        ) -> pd.DataFrame:
        """
        Run a query string and return results in a pandas DataFrame
        """
        cursor = None
        try:
            cursor = self.connection.cursor()
            query_return = cursor.execute(query, params).fetch_pandas_all()
            return query_return
        except Exception as e:
            self._error_handler(e)
        finally:
            if cursor:
                cursor.close()

    def _execute_string(
            self,
            query: str,
            ignore_results: bool = False
        ) -> List[tuple]:
        """
        Execute a query string against the Data Warehouse connection and fetch all results
        """
        query_returns = []
        cursor = None
        try:
            for cursor in self.connection.execute_string(query, return_cursors=(not ignore_results)):
                for query_return in cursor:
                    query_returns.append(query_return)
            return query_returns
        except Exception as e:
            self._error_handler(e)
        finally:
            if cursor:
                cursor.close()



"""
RasgoQL package entrypoint
"""
from typing import Union

import webbrowser

from rasgoql.data import (
    DWCredentials,
    BigQueryCredentials, BigQueryDataWarehouse,
    SnowflakeCredentials, SnowflakeDataWarehouse,
    PostgresCredentials, PostgresDataWarehouse
)
from rasgoql.errors import ParameterException
from rasgoql.main import RasgoQL
from rasgoql.version import __version__

__all__ = [
    '__version__',
    'docs',
    'connect',
]

DW_MAP = {
    'snowflake': SnowflakeDataWarehouse,
    'bigquery': BigQueryDataWarehouse,
    'postgresql': PostgresDataWarehouse,
}


def docs():
    """
    Open Rasgo Docs page
    """
    url = 'https://docs.rasgoql.com'
    webbrowser.open(url)
    print(url)

def connect(
        credentials: DWCredentials
    ):
    """
    Return a RasgoQL object connected to a Data Warehouse
    """
    return RasgoQL(
        connection=DW_MAP[credentials.dw_type],
        credentials=credentials.to_dict()
    )



"""
Primitive Classes
"""
import logging
from typing import Callable, List

import pandas as pd
import rasgotransforms as rtx

from rasgoql.errors import ParameterException
from rasgoql.utils.decorators import (
    beta, require_dw,
    require_materialized, require_transforms
)
from rasgoql.utils.sql import (
    parse_fqtn, make_namespace_from_fqtn,
    random_table_name, validate_fqtn
)
from rasgoql.primitives.enums import (
    check_render_method, check_table_type, check_write_table_type,
    TableType, TableState
)
from rasgoql.primitives.rendering import (
    assemble_cte_chain, assemble_view_chain, create_dbt_files,
    _gen_udt_func_docstring, _gen_udt_func_signature
)

logging.basicConfig()
ds_logger = logging.getLogger('Dataset')
ds_logger.setLevel(logging.INFO)
chn_logger = logging.getLogger('SQLChain')
chn_logger.setLevel(logging.INFO)


class TransformableClass:
    """
    Class to attach Rasgo transform methods to other classes
    """
    def __init__(
            self,
            dw: 'DataWarehouse'
        ):
        self._dw = dw
        self._transform_sync()

    def _create_aliased_function(
            self,
            transform: 'TransformTemplate'
        ) -> Callable:
        """
        Returns a new function to dynamically attach to a class on init
        """
        def f(*arg, **kwargs) -> 'SQLChain':
            return self.transform(name=transform.name, *arg, **kwargs)

        f.__name__ = transform.name
        f.__signature__ = _gen_udt_func_signature(f, transform)
        f.__doc__ = _gen_udt_func_docstring(transform)
        return f

    def _transform_sync(self):
        """
        Gather available transforms and create aliased functions for each
        """
        for transform in rtx.serve_rasgo_transform_templates(self._dw.dw_type):
            f = self._create_aliased_function(transform)
            setattr(self, transform.name, f)

    @require_dw
    def transform(
            self,
            name: str,
            arguments: dict = None,
            output_alias: str = None,
            **kwargs
        ) -> 'SQLChain':
        """
        Apply a Rasgo transform and return a SQLChain
        """
        source_table = ''
        transforms = []
        # If we're transforming a Dataset
        if isinstance(self, Dataset):
            entry_table = self
            source_table = self.fqtn
            namespace = self._dw.default_namespace
        # If we're transforming a Chain
        if isinstance(self, SQLChain):
            entry_table = self.entry_table
            source_table = self.output_table.fqtn
            namespace = self.namespace
            # If we're transforming a Chain with existing transforms
            if self.transforms:
                source_table = self.output_table.table_name
                transforms = transforms + self.transforms
        arguments = arguments if arguments else {}
        arguments.update(kwargs)
        arguments['source_table'] = source_table
        new_transform = Transform(
            name,
            arguments,
            namespace,
            output_alias,
            self._dw
        )
        transforms.append(new_transform)
        return SQLChain(entry_table, namespace, transforms, self._dw)


class Dataset(TransformableClass):
    """
    Reference to a table or view in the Data Warehouse
    """
    def __init__(
            self,
            fqtn: str,
            dw: 'DataWarehouse' = None
        ):
        super().__init__(dw)
        try:
            self.fqtn: str = validate_fqtn(fqtn)
            self.table_name: str = parse_fqtn(fqtn)[2]
            self.namespace: str = make_namespace_from_fqtn(fqtn)
            self._dw._validate_namespace(self.namespace)
        except ValueError:
            raise ParameterException("Must pass in a valid 'fqtn' parameter to create a Dataset")
        self.table_type: str = TableType.UNKNOWN.value
        self.table_state: str = TableState.UNKNOWN.value
        self.is_rasgo: bool = False
        self._dw_sync()

    def __repr__(self) -> str:
        return f"Dataset(fqtn={self.fqtn}, " \
               f"type={self.table_type}, " \
               f"state={self.table_state})"

    @require_dw
    def _dw_sync(self):
        """
        Synchronize status with Cloud DataWarehouse
        """
        obj_exists, is_rasgo_obj, obj_type = self._dw.get_object_details(self.fqtn)
        if obj_exists:
            self.table_state = TableState.IN_DW.value
            self.table_type = check_table_type(obj_type)
            self.is_rasgo = is_rasgo_obj
        else:
            self.table_state = TableState.IN_MEMORY.value

    @require_dw
    def get_schema(self) -> dict:
        """
        Return the schema of this table
        """
        return self._dw.get_schema(self.fqtn)

    @require_dw
    @require_materialized
    def preview(self) -> pd.DataFrame:
        """
        Return a pandas DataFrame of top 10 rows
        """
        return self._dw.preview(f'SELECT * FROM {self.fqtn}')

    @require_dw
    def sql(self) -> dict:
        """
        Return the ddl to create this table
        """
        return self._dw.get_ddl(self.fqtn)

    @require_dw
    @require_materialized
    def to_df(self) -> pd.DataFrame:
        """
        Return a pandas DataFrame of the entire table
        """
        return self._dw.execute_query(f"SELECT * FROM {self.fqtn}", response='df')


class TransformTemplate:
    """
    Reference to a Rasgo transform template
    """
    def __init__(
            self,
            name: str,
            arguments: List[dict],
            source_code: str,
            transform_type: str,
            description: str = None
        ):
        self.name = name
        self.arguments = arguments
        self.source_code = source_code
        self.transform_type = transform_type
        self.description = description

    def __repr__(self) -> str:
        arg_str = ', '.join(f'{arg.get("name")}: {arg.get("type")}' for arg in self.arguments)
        return f"RasgoTemplate: {self.name}({arg_str})"

    def define(self) -> str:
        """
        Return a pretty string definition of this Transform
        """
        pretty_string = f'''{self.transform_type.title()} Transform: {self.name}
        Description: {self.description}
        Arguments: {self.arguments}
        SourceCode: {self.source_code}
        '''
        return pretty_string


class Transform:
    """
    Reference to a Transform
    """
    def __init__(
            self,
            name: str,
            arguments: dict,
            namespace: str,
            output_alias: str = None,
            dw: 'DataWarehouse' = None
        ):
        self._dw = dw

        self.arguments = arguments
        self.name = name
        self.namespace = namespace
        self.output_alias = output_alias or random_table_name()
        self.source_table = arguments.get("source_table")

    def __repr__(self) -> str:
        return f"Transform({self.output_alias}: {self.name})"

    @property
    def fqtn(self) -> str:
        """
        Returns the fully qualified table name the transform would create if saved
        """
        return f'{self.namespace}.{self.output_alias}'


class SQLChain(TransformableClass):
    """
    Reference to one or more Transforms
    """

    def __init__(
            self,
            entry_table: Dataset,
            namespace: str,
            transforms: List[Transform] = None,
            dw: 'DataWarehouse' = None
        ) -> None:
        super().__init__(dw)
        self.entry_table = entry_table
        self.transforms = transforms
        self.namespace = namespace

    def __repr__(self) -> str:
        transform_count = len(self.transforms) if self.transforms else 0
        return f"SQLChain({self.entry_table.fqtn} + {transform_count} transforms)"

    def change_namespace(
            self,
            new_namespace: str
        ):
        """
        Re-sets the class's namespace
        """
        self._dw.change_namespace(new_namespace)
        self.namespace = new_namespace
        for t in self.transforms:
            t.namespace = new_namespace

    @property
    def fqtn(self) -> str:
        """
        Returns the fully qualified table name this SQLChain would create
        if saved in current state

        NOTE: This property will be dynamic until the Chain is finally saved
        """
        return self.output_table.fqtn

    @require_dw
    def get_schema(self) -> dict:
        """
        Return the table schema this SQLChain would create if saved in current state

        NOTE: This property will be dynamic until the Chain is finally saved
        """
        return self._dw.get_schema(self.fqtn, self.sql())

    @property
    def output_table(self) -> Dataset:
        """
        Returns the Dataset this SQLChain would create if saved in current state

        NOTE: This property will be dynamic until the Chain is finally saved
        """
        if self.ternimal_transform:
            return Dataset(self.ternimal_transform.fqtn, self._dw)
        return self.entry_table

    @require_dw
    @require_transforms
    def preview(self) -> pd.DataFrame:
        """
        Returns the top 10 rows of data into a pandas DataFrame
        """
        return self._dw.preview(self.sql())

    @require_dw
    def sql(
            self,
            render_method: str = 'SELECT'
        ) -> str:
        """
        Returns the SQL to build this Transform Chain

        render_method: str: ['SELECT', 'TABLE', 'VIEW', 'VIEW CHAIN']
        """
        render_method = check_render_method(render_method)
        if not self.transforms:
            return f"SELECT * FROM {self.output_table.fqtn}"
        if render_method == 'VIEWS':
            return assemble_view_chain(self.transforms)
        return assemble_cte_chain(
            self.transforms,
            render_method if render_method in ['TABLE', 'VIEW'] else None
        )

    @require_dw
    @require_transforms
    def save(
            self,
            table_name: str = None,
            table_type: str = 'view',
            overwrite: bool = False
        ):
        """
        Materializes this Transform Chain into SQL objects
        """
        table_type = check_write_table_type(table_type)
        table_name = table_name or self.output_table.fqtn
        new_table = self._dw.create(
            self.sql(),
            table_name,
            table_type,
            overwrite
        )
        return Dataset(new_table, self._dw)

    @property
    def ternimal_transform(self) -> Transform:
        """
        Returns the last transform in this chain
        """
        if self.transforms:
            return self.transforms[-1]
        return None

    @beta
    def to_dbt(
            self,
            output_directory: str = None,
            file_name: str = None,
            config_args: dict = None,
            include_schema: bool = False
        ) -> str:
        """
        Saves a new model.sql file to your dbt models directory

        Params:
        `output_directory`: str:
            directory to save model file
            defaults to current working dir
        `file_name`: str:
            defaults to {output_alias}.sql of SQLChain
        `include_schema`: bool:
            Include a schema.yml file
        `config_args`: dict:
            key value pair of
            dbt [config values](https://docs.getdbt.com/reference/model-configs)
        """
        try:
            schema = self.get_schema()
        except:
            if include_schema:
                chn_logger.warning(
                    'Unexpected error generating the schema of this SQLChain. '
                    'Your model.sql file will be generated without a schema.yml file. '
                    'This is most likely a syntax issue in your SQLChain or existing view. '
                    'Consider running your_chn.sql() to check the syntax and/or '
                    'your_chn.save() to update the view definition in your Data Warehouse.'
                )
            schema = []
        return create_dbt_files(
            self.transforms,
            schema,
            output_directory,
            file_name,
            config_args,
            include_schema
        )

    def to_df(self) -> pd.DataFrame:
        """
        Returns data into a pandas DataFrame
        """
        return self._dw.execute_query(self.sql(), response='df')



"""
Transform rendering methods
"""
import functools
import inspect
from itertools import combinations, permutations, product
import re
import os
from typing import Callable, Dict, List, Tuple, Optional

import jinja2
import pandas as pd
import rasgotransforms as rtx

from rasgoql.errors import TransformRenderingError
from rasgoql.primitives.enums import check_write_table_type
from rasgoql.utils.dbt import save_model_file

RUN_QUERY_LIMIT = 100
JINJA_ENV = jinja2.Environment(extensions=['jinja2.ext.do', 'jinja2.ext.loopcontrols'])


def assemble_cte_chain(
        transforms: List['Transform'],
        table_type: str = None
    ):
    """
    Returns a nested CTE statement to render input Transforms
    """
    create_stmt, final_select = '', ''

    # Handle single transform chains
    if len(transforms) == 1:
        t = transforms[0]
        create_stmt = _set_create_statement(table_type, t.fqtn)
        final_select = generate_transform_sql(
            t.name,
            t.arguments,
            t.source_table,
            None,
            t._dw
        )
        return create_stmt + final_select

    # Handle multi-transform chains
    t_list = []
    running_sql = None
    for t in transforms:
        t_sql = generate_transform_sql(
            t.name,
            t.arguments,
            t.source_table,
            running_sql,
            t._dw
        )

        # Set final select & create statements from terminal transform
        if t == transforms[-1]:
            final_select = _set_final_select_statement(len(transforms), t_sql)
            create_stmt = _set_create_statement(table_type, t.fqtn)
        else:
            running_sql = _construct_running_sql(t_list, t_sql)
            t_cte_str = f'{t.output_alias} AS (\n{t_sql}\n) '
            t_list.append(t_cte_str)
    return create_stmt + 'WITH ' + ', \n'.join(t_list) + final_select

def assemble_view_chain(
        transforms: List['Transform']
    ):
    """
    Returns a nested view statement to render input Transforms
    """
    view_list = []
    cte_list = []
    running_sql = None
    for t in transforms:
        t_sql = generate_transform_sql(
            t.name,
            t.arguments,
            t.source_table,
            running_sql,
            t._dw
        )
        running_sql = _construct_running_sql(cte_list, t_sql)
        t_cte_str = f'{t.output_alias} AS (\n{t_sql}\n) '
        t_view_str = f'CREATE OR REPLACE VIEW {t.fqtn} AS {t_sql};'
        cte_list.append(t_cte_str)
        view_list.append(t_view_str)
    return '\n'.join(view_list)

def create_dbt_files(
        transforms: List['Transform'],
        schema: List[Tuple[str, str]],
        output_directory: str = None,
        file_name: str = None,
        config_args: dict = None,
        include_schema: bool = False
) -> str:
    """
    Saves a dbt_project.yml and model.sql files to a directory
    """
    output_directory = output_directory or os.getcwd()
    file_name = file_name or f'{transforms[-1].output_alias}.sql'
    return save_model_file(
        sql_definition=assemble_cte_chain(transforms),
        output_directory=output_directory,
        file_name=file_name,
        config_args=config_args,
        include_schema=include_schema,
        schema=schema
    )

def generate_transform_sql(
        name: str,
        arguments: dict,
        source_table: str = None,
        running_sql: str = None,
        dw: 'DataWarehouse' = None
    ) -> str:
    """
    Returns the SQL for a Transform with applied arguments
    """
    templates = rtx.serve_rasgo_transform_templates(dw.dw_type)
    udt: 'TransformTemplate' = [t for t in templates if t.name == name][0]
    if not udt:
        raise TransformRenderingError(f'Cannot find a transform named {name}')
    try:
        source_code = udt.source_code
        if udt.name == 'apply':
            source_code = arguments.pop('sql')
            if not source_code:
                raise TransformRenderingError(f'Custom transform {udt.name} must provide ' \
                                               'the "sql" argument with template to apply')
        return render_template(name, source_code, arguments, source_table, running_sql, dw)
    except Exception as e:
        raise TransformRenderingError(e) from e

def render_template(
        name: str,
        source_code: str,
        arguments: dict,
        source_table: str,
        running_sql: str,
        dw: 'DataWarehouse'
    ) -> str:
    """
    Returns a SQL statement generated by applying arguments to a Jinja template
    """
    template_fns = _source_code_functions(dw, source_table, running_sql)
    template = JINJA_ENV.from_string(source_code)
    rendered = template.render(**arguments, **template_fns)
    if not rendered:
        raise TransformRenderingError(f'Rendering of transform {name} produced an empty string. ' \
                                       'Please check template arguments and try again.')
    return rendered

def _cleanse_template_symbol(
        symbol: str
    ) -> str:
    """
    Extra verbose function for clarity

    remove double quotes
    replace spaces and dashes with underscores
    cast to upper case
    delete anything that is not letters, numbers, or underscores
    if first character is a number, add an underscore to the beginning
    """
    symbol = str(symbol)
    symbol = symbol.strip()
    symbol = symbol.replace(' ', '_').replace('-', '_')
    symbol = symbol.upper()
    symbol = re.sub('[^A-Z0-9_]+', '', symbol)
    symbol = '_'+symbol if symbol[0].isdecimal() or not symbol else symbol
    return symbol

def _collapse_cte(
        sql: str
    ) -> str:
    """
    Returns a collapsed CTE if sql is a CTE, or original sql
    """
    if sql.upper().startswith('WITH'):
        return re.sub(r'^(WITH)\s', r', ', sql, 1, flags=re.IGNORECASE)
    return sql

def _construct_running_sql(
        cte_list: List[str],
        sql: str
    ) -> str:
    """
    Constructs and returns a running sql statement
    """
    # running_sql tracks the CTE chain up to this transform,
    # this is consumed by the next transform if it needs to
    # access the chain's data mid-state (e.g. confirming columns exits)
    if len(cte_list) == 0:
        return sql
    return 'WITH ' + ', \n'.join(cte_list) + _collapse_cte(sql)

def _gen_udt_func_docstring(
        transform: 'TransformTemplate'
    ) -> str:
    """
    Generate and return a docstring for a transform func
    with transform description, args, and return specified.
    """
    docstring = f"\n{transform.description}"

    docstring = f"{docstring}\n  Args:"
    for t_arg in transform.arguments:
        docstring = f"{docstring}\n    {t_arg.get('name')}: {t_arg.get('description')}"
    docstring = f"{docstring}\n    operation_name: Name to set for the operation"

    docstring = f"{docstring}\n\n  Returns:\n    Returns an new dataset with the referenced " \
                f"{transform.name!r} added to this dataset's definition"
    return docstring

def _gen_udt_func_signature(
        udt_func: Callable,
        transform: 'TransformTemplate'
    ) -> inspect.Signature:
    """
    Creates and returns a UDT param signature.

    This is shown documentation for the parameters when hitting shift tab in a notebook
    """
    sig = inspect.signature(udt_func)

    udt_params = []
    for t_arg in transform.arguments:
        p = inspect.Parameter(name=t_arg.get('name'), kind=inspect.Parameter.KEYWORD_ONLY)
        udt_params.append(p)

    op_name_param = inspect.Parameter(
        name='operation_name',
        kind=inspect.Parameter.KEYWORD_ONLY,
        annotation=Optional[str],
        default=None
    )
    udt_params.append(op_name_param)
    return sig.replace(parameters=udt_params)

def _raise_exception(
        message: str
    ) -> None:
    """
    Raise an exception to return to users of a template
    """
    raise TransformRenderingError(message)

def _run_query(
        query: str,
        source_table: str = None,
        running_sql: str = None,
        dw: 'DataWarehouse' = None
    ) -> pd.DataFrame:
    """
    Jinja Func to materialize a chain as a temporary view before running a query
    """
    try:
        if running_sql:
            create_sql = f"CREATE OR REPLACE VIEW {source_table} AS {running_sql} LIMIT {RUN_QUERY_LIMIT}"
            dw.execute_query(create_sql, response='none', acknowledge_risk=True)
        return dw.execute_query(query, response='df', acknowledge_risk=True)
    except Exception as e:
        raise
    finally:
        if running_sql:
            drop_sql = f"DROP VIEW IF EXISTS {source_table}"
            dw.execute_query(drop_sql, response='none', acknowledge_risk=True)

def _set_create_statement(
        table_type: str,
        fqtn: str,
    ) -> str:
    """
    Returns a create statement or a blank string
    """
    if table_type:
        table_type = check_write_table_type(table_type)
        return f'CREATE OR REPLACE {table_type} {fqtn} AS \n'
    return ''

def _set_final_select_statement(
        transform_count: int,
        sql: str,
    ) -> str:
    """
    Returns a create statement or a blank string
    """
    if transform_count > 1:
        return _collapse_cte(sql)
    return sql

def _source_code_functions(
        dw: 'DataWarehouse',
        source_table: str = None,
        running_sql: str = None
    ) -> Dict[str, Callable]:
    """
    Get custom functions to add to the template parser
    """
    return {
        "run_query": functools.partial(
            _run_query,
            dw=dw,
            source_table=source_table,
            running_sql=running_sql
        ),
        "cleanse_name": _cleanse_template_symbol,
        "raise_exception": _raise_exception,
        "itertools": {
            "combinations": combinations,
            "permutations": permutations,
            "product": product
        }
    }
